/* Microprocessors Based Design
 * Asynchronous Web Server
 *
 * Calin Cruceru <calin.cruceru@stud.acs.upb.ro>, 335CB
 *
 * Module which implements the byte supplier logic.
 */

/* Project */
#include <aws.h> /* AWS_RNG_DEVICE */

#include "bytes_supply.h"
#include "debug.h"
#include "error_handling.h"

/* Syscalls */
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

/* mbedtls */
#include <mbedtls/entropy.h>
#include <mbedtls/ctr_drbg.h>

/* Static data */
static int hwrng_fd;
static mbedtls_entropy_context entropy_ctx;
static mbedtls_ctr_drbg_context drbg_ctx;

static int entropy_source_fun(void __attribute__((unused)) *data, /* NULL */
			      unsigned char *output, size_t len, size_t *olen)
{
	ssize_t bytes_read;

	bytes_read = read(hwrng_fd, output, len);
	DIE(bytes_read < 0, "read");

	*olen = bytes_read;
	return 0;
}

int get_random_bytes(unsigned char *output, size_t output_len)
{
	int rc;

	rc = mbedtls_ctr_drbg_random(&drbg_ctx, output, output_len);
	DIE(rc != 0, "mbedtls_ctr_drbg_random");

	return 0;
}

/* Constructor function.  It runs at load time. */
__attribute__((constructor))
static void init_bytes_supply(void)
{
	int rc;

	/* Init TRNG file descriptor */
	hwrng_fd = open(AWS_RNG_DEVICE, O_RDONLY | O_NOCTTY);
	DIE(hwrng_fd < 0, "open");

	/* Init entropy accumulator */
	mbedtls_entropy_init(&entropy_ctx);
	rc = mbedtls_entropy_add_source(&entropy_ctx, &entropy_source_fun,
					NULL, MBEDTLS_CTR_DRBG_ENTROPY_LEN,
					MBEDTLS_ENTROPY_SOURCE_STRONG);
	DIE(rc != 0, "mbed_entropy_add_source");

	/* Init DRBG */
	mbedtls_ctr_drbg_init(&drbg_ctx);
	rc = mbedtls_ctr_drbg_seed(&drbg_ctx, &mbedtls_entropy_func,
				   &entropy_ctx, NULL, 0);
	DIE(rc != 0, "mbedtls_ctr_drbg_seed");
	mbedtls_ctr_drbg_set_prediction_resistance(&drbg_ctx,
						   MBEDTLS_CTR_DRBG_PR_ON);

	dlog(LOG_INFO, "Finished initializing bytes supply!\n");
}

