/* Microprocessors Based Design
 * Asynchronous Web Server
 *
 * Calin Cruceru <calin.cruceru@stud.acs.upb.ro>, 335CB
 *
 * The `connection` type which encapsulates all the data held during a
 * connection in our AWS.
 */

#ifndef CONNECTION_H_
#define CONNECTION_H_ 1

#include <aws.h> /* RAND_BYTES_PER_REQUEST */
#include <http_parser.h> /* HTTP_MAX_HEADER_SIZE */

/* A connection can be in any of the states which appear in the following
 * automata:
 *
 *				+----------------+
 *				| Receiving Data |
 *				+----------------+
 *					|
 *					|
 *				+---------------------+
 *				| Finished Recv. Data |
 *				+---------------------+
 *					|
 *					|
 *					|
 *	(invalid request)		|		(bits request)
 *		+-----------------------------------------------+
 *		|						|
 *		|						|
 *	+---------------+				+----------------+
 *	| Sending Error |				| Send rand bits |
 *	+---------------+				+----------------+
 *		|						|
 *		|						|
 *		+-----------------------------------------------+
 *					|
 *					|
 *				+------------------+
 *				| Connection Ended |
 *				+------------------+
 *
 * If an error (other than an invalid file request, which is handled by
 * replying with an HTTP 404 message) happens in one of the states, such as the
 * client shuts down the connection, etc, we release the resources acquired
 * until that point and we transition to STATE_CONNECTION_ERROR.
 *
 *	|			|		...			|
 *	|			|		...			|
 *	-----------------------------------------------------------------
 *					|
 *					|
 *				+------------------+
 *				| Connection Error |
 *				+------------------+
 */
enum connection_state {
	/* Special state used by the aws for accepting new connections.
	 * Only one connection should be in this state.
	 */
	STATE_LISTENING,

	STATE_RECEIVING_DATA,
	STATE_FINISHED_RECEIVING_DATA,
	STATE_SENDING_ERROR,
	STATE_SENDING_RAND_BITS,
	STATE_CONNECTION_ENDED,

	/* Also a special state used to signal a (nonfatal) error during the
	 * communication.
	 */
	STATE_CONNECTION_ERROR,
};

#define AWS_HTTP_MAX_BODY_SIZE 65536
#define AWS_HTTP_MAX_MSG_SIZE (HTTP_MAX_HEADER_SIZE + AWS_HTTP_MAX_BODY_SIZE)

struct connection {
	/* The current state of the connection */
	enum connection_state state;

	/* The socket file descriptor used to communicate with the client */
	int sockfd;

	/* The input buffer used to store the request from the client. */
	char inbuf[AWS_HTTP_MAX_MSG_SIZE];
	off_t inoff;

	/* The output buffer used to store the HTTP reply */
	char outbuf[HTTP_MAX_HEADER_SIZE + 3 * RAND_BYTES_PER_REQUEST];
	off_t outoff;
	size_t outsz;
};

/* Allocation and deallocation functions for the `connection` data type */
extern struct connection *conn_new(int sockfd);
extern void conn_free(struct connection *conn);

#endif /* CONNECTION_H_ */

