/* Microprocessors Based Design
 * Asynchronous Web Server
 *
 * Calin Cruceru <calin.cruceru@stud.acs.upb.ro>, 335CB
 *
 * Asynchronous Web Server
 */

/* Project */
#include "bytes_supply.h"
#include "connection.h"
#include "debug.h"
#include "error_handling.h"
#include "http_util.h"
#include "sock_util.h"
#include "w_epoll.h"

/* Linux */
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/epoll.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

/* C */
#include <string.h>
#include <stdint.h>

/* Static data */
static struct connection listening_conn;
static int epollfd;

static size_t handle_new_connections(void)
{
	int sockfd;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	struct sockaddr_in addr;
	struct connection *conn;
	size_t connection_count = 0;
	int rc;

	for (;;) {
		/* Accept a new connection */
		sockfd = accept4(listening_conn.sockfd,
				 (struct sockaddr *) &addr,
				 &addrlen,
				 SOCK_NONBLOCK);

		switch (sockfd) {
		case -1: /* errno should be either EAGAIN or EWOULDBLOCK */
			DIE(!(errno == EAGAIN || errno == EWOULDBLOCK),
			    "accept4");
			goto _exit;

		default:
			dlog(LOG_INFO, "Accepted connection from %s:%d\n",
			     inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));

			/* Create a new connection and add it to EPOLL */
			conn = conn_new(sockfd);
			rc = w_epoll_add_ptr_in(epollfd, sockfd, conn);
			DIE(rc < 0, "w_epoll_add_ptr_in");

			/* Increment the number of new connections started */
			++connection_count;
		}
	}

_exit:
	return connection_count;
}

static void close_connection(struct connection *conn)
{
	dlog(LOG_DEBUG, "Closing connection...\n");

	/* Remove it from EPOLL */
	DIE(w_epoll_remove_ptr(epollfd, conn->sockfd, conn) < 0,
	    "w_epoll_remove_ptr");

	/* Free its resources */
	conn_free(conn);
}

static enum connection_state advance_recv_data(struct connection *conn)
{
	int rc;

	/* Remove it from POLLIN and add it to POLLOUT */
	rc = w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);
	DIE(rc < 0, "w_epoll_update_ptr_out");

	conn->state = STATE_FINISHED_RECEIVING_DATA;
	return conn->state;
}

static void add_http_200_header(struct connection *conn)
{
	char http_header[HTTP_MAX_HEADER_SIZE];
	int http_header_len;

	/* Write the header into a temporary buffer so we can take its size */
	http_header_len = sprintf(
		/* Output buffer */
		http_header,

		/* Our actual header */
		"HTTP/1.1 200 OK\r\n"
		"Content-Type: text/plain\r\n"
		"Content-Length: %zu\r\n"
		"Connection: close\r\n"
		"Date: Thu, 26 May 2016 02:50:38 +0000\r\n"
		"\r\n",

		/* Args referenced in the string literal from above */
		conn->outsz
	);

	memmove(conn->outbuf + http_header_len, conn->outbuf, conn->outsz);
	memcpy(conn->outbuf, http_header, http_header_len);
	conn->outsz += http_header_len;
}

static void add_head_message(struct connection *conn)
{
	conn->outsz += sprintf(
		/* Output buffer */
		conn->outbuf + conn->outsz,

		/* Head message */
		"Here are %d random bytes.  Down this page you can find more "
		"about this TRNG.\n\n",

		/* Args referenced in the string literal from above */
		RAND_BYTES_PER_REQUEST
	);
}

static void add_tail_message(struct connection *conn)
{
	conn->outsz += sprintf(
		conn->outbuf + conn->outsz,

		/* Tail message */
		"\n\nThe following pipeline is used to generate the stream "
		"from above:\nTODO\n"
	);
}

/* AWS_BYTES_PER_LINE is 26, which is 80 divided by 3 - the characters needed
 * to display a hex byte plus a space following it
 */
#define AWS_BYTES_PER_LINE 26

static void advance_recv_to_send(struct connection *conn)
{
	unsigned char buffer[RAND_BYTES_PER_REQUEST];
	int i, rc;

	rc = get_random_bytes(buffer, sizeof(buffer));
	DIE(rc < 0, "get_random_bytes");

	/* Prepare HTTP 200 reply: initialize the output buffer */
	memset(conn->outbuf, 0, sizeof(conn->outbuf));
	conn->outsz = 0;
	conn->outoff = 0;

	/* Add the following:
	 *	* head message
	 *	* random bytes displayed in hexadecimal
	 *	* tail message
	 */
	add_head_message(conn);

	for (i = 0; i < RAND_BYTES_PER_REQUEST; ++i) {
		if (i && i % AWS_BYTES_PER_LINE == 0)
			conn->outsz += sprintf(conn->outbuf + conn->outsz,
					       "\n");

		conn->outsz += sprintf(conn->outbuf + conn->outsz, "%.2x ",
				       buffer[i]);
	}

	add_tail_message(conn);

	/* Now that we know the size of the http reply body, we can add the
	 * header as well.
	 */
	add_http_200_header(conn);

	conn->state = STATE_SENDING_RAND_BITS;
}

static void advance_recv_to_err(struct connection *conn)
{
	static const char *http_404
		= "HTTP/1.1 404 Not Found\r\n"
		  "Content-Type: text/plain\r\n"
		  "Content-Length: 33\r\n"
		  "Connection: close\r\n"
		  "Date: Thu, 26 May 2016 02:50:38 +0000\r\n"
		  "\r\n"
		  "The server only replies to /bytes!";

	dlog(LOG_DEBUG, "HTTP GET requests invalid file\n");

	/* Add 404 HTTP header */
	memcpy(conn->outbuf, http_404, strlen(http_404));
	conn->outoff = 0;
	conn->outsz = strlen(http_404);

	conn->state = STATE_SENDING_ERROR;
}

static enum connection_state advance_finished_recv(struct connection *conn)
{
	char path[MAX_PATH_SIZE];

	/* Extract the path requested by the client */
	extract_path(conn->inbuf, path);
	dlog(LOG_DEBUG, "The requested file is %s\n", path);

	/* If it matches our only accepted requested file */
	if (!strcmp(path, AWS_REL_BITS_FILE)) {
		advance_recv_to_send(conn);
	} else {
		advance_recv_to_err(conn);
	}

	return conn->state;
}

static enum connection_state advance_send_rand_bits(struct connection *conn)
{
	conn->state = STATE_CONNECTION_ENDED;
	return conn->state;
}

static enum connection_state advance_send_error(struct connection *conn)
{
	conn->state = STATE_CONNECTION_ENDED;
	return conn->state;
}

static void handle_sock_event(struct connection *conn)
{
	enum connection_state next_state;
	ssize_t ret;

	switch (conn->state) {
	case STATE_RECEIVING_DATA:
		ret = recv_nonblock(conn->sockfd, conn->inbuf, &conn->inoff,
				    HTTP_MAX_HEADER_SIZE - conn->inoff);
		if (ret < 0) {
			conn->state = STATE_CONNECTION_ERROR;
			goto _close_connection;
		}

		if (strstr(conn->inbuf, "\r\n\r\n")) {
			next_state = advance_recv_data(conn);

			switch (next_state) {
			case STATE_FINISHED_RECEIVING_DATA:
				/* Nothing to do */
				break;

			case STATE_CONNECTION_ERROR:
				goto _close_connection;

			default:
				DIE(1, "automata invariant");
			}
		}
		break;

	case STATE_FINISHED_RECEIVING_DATA:
		next_state = advance_finished_recv(conn);

		switch (next_state) {
		case STATE_SENDING_RAND_BITS:
		case STATE_SENDING_ERROR:
			/* We call this so that we don't lose the
			 * edge-triggered epoll signal.
			 */
			handle_sock_event(conn);
			break;

		case STATE_CONNECTION_ERROR:
			goto _close_connection;

		default:
			DIE(1, "automata invariant");
		}
		break;

	case STATE_SENDING_ERROR:
		ret = send_nonblock(conn->sockfd, conn->outbuf, &conn->outoff,
				    conn->outsz - conn->outoff);
		if (ret < 0) {
			conn->state = STATE_CONNECTION_ERROR;
			goto _close_connection;
		}

		if ((off_t) conn->outsz == conn->outoff) {
			next_state = advance_send_error(conn);

			switch (next_state) {
			case STATE_CONNECTION_ENDED:
			case STATE_CONNECTION_ERROR:
				goto _close_connection;

			default:
				DIE(1, "automata invariant");
			}
		}
		break;

	case STATE_SENDING_RAND_BITS:
		ret = send_nonblock(conn->sockfd, conn->outbuf, &conn->outoff,
				    conn->outsz - conn->outoff);
		if (ret < 0) {
			conn->state = STATE_CONNECTION_ERROR;
			goto _close_connection;
		}

		if ((off_t) conn->outsz == conn->outoff) {
			next_state = advance_send_rand_bits(conn);

			switch (next_state) {
			case STATE_CONNECTION_ENDED:
			case STATE_CONNECTION_ERROR:
				goto _close_connection;

			default:
				DIE(1, "automata invariant");
			}
		}
		break;

	default:
		DIE(1, "automata invariant");
	}

	return;

_close_connection:
	close_connection(conn);
}

int main(void)
{
	int rc;

	/* Initialize EPOLL */
	epollfd = w_epoll_create();
	DIE(epollfd < 0, "w_epoll_create");

	/* Initialize the special 'listening connection' */
	listening_conn.state = STATE_LISTENING;
	listening_conn.sockfd = tcp_create_listener(AWS_LISTEN_PORT,
						    DEFAULT_LISTEN_BACKLOG);

	/* Add it to EPOLL */
	rc = w_epoll_add_ptr_in(epollfd, listening_conn.sockfd,
				&listening_conn);
	DIE(rc < 0, "w_epoll_add_ptr_in");

	dlog(LOG_INFO, "AWS is listening on port %d\n", AWS_LISTEN_PORT);

	/* Server main loop */
	for (;;) {
		struct epoll_event epoll_ev;
		struct connection *conn;

		/* Wait for events */
		rc = w_epoll_wait_infinite(epollfd, &epoll_ev);
		DIE(rc < 0, "w_epoll_wait_infinite");

		conn = epoll_ev.data.ptr;
		dlog(LOG_INFO, "Processing connection with state %d\n",
		     conn->state);

		if (conn->state == STATE_LISTENING) {
			/* It's a new connection */
			size_t count;

			dlog(LOG_INFO, "New connections awaiting\n");
			count = handle_new_connections();

			dlog(LOG_INFO, "%zu connections just started\n", count);
		} else if ((epoll_ev.events & EPOLLIN) ||
			   (epoll_ev.events & EPOLLOUT)) {
			/* A socket is ready to be written or read.
			 * The automata deals with both of them.
			 */
			handle_sock_event(conn);
		}
	}

	return 0;
}

